#|[
Replacement of car, cdr and consorts by one operator, walk,
of integer arguments. (walk -1) is car, (walk 1) cdr, (walk k) is
(iterate cdr k) and (walk -k) (iterate car (- k)).
These operaters can (and will) be composed, so, for example,
(walk -1 2) is (o (walk -1) (walk 2)), in other words caddr.

The more arguments you give to walk, the deeper you can dig into a
tree. If you like, you can interprete the arguments of walk as a path
for traverseing the tree. For compatibility with the c*r routines,
you interprete the arguments from walk from right to left. It might be
more convenient, to do it left to right; walk* does this.
]|#

(module tree-walkers (tree-walkers iterate walk walk*)
  (import scheme
          (only (chicken base) o error)
          (only (chicken pretty-print) pp))

(define (tree-walkers . args)
  (cond
    ((null? args)
     '(iterate walk walk*))
    ((null? (cdr args))
     (case (car args)
       ((iterate)
        (pp (list
              '(iterate fn k start)
              '(iterate fn k)
              "apply fn successively k times to start")))
       ((walk)
        (pp
          (list '(walk . ks)
              "transform the integers in ks from right to left"
              "to corresponding car's and cdr's"
              "and apply them to a tree, i.e. a nested pseudolist")))
       ((walk*)
        (pp
          (list '(walk* . ks)
                "same as walk, except the k's are transformed"
                "left to right")))
       (else
         (pp "argument not understood"))))))

(define (iterate fn times . rest)
  (cond
    ((null? rest)
     (lambda (arg)
       (let loop ((k times))
         (if (zero? k)
           arg
           (fn (loop (- k 1)))))))
    ((null? (cdr rest))
     ((iterate fn times) (car rest)))
    (else (error 'iterate "too many arguments" rest))))
     
(define (walk . ks)
  (cond
    ((null? ks) (lambda (x) x))
    ((null? (cdr ks))
     (let ((k (car ks)))
       (cond
         ((>= k 0)
          (iterate cdr k))
         ((< k 0)
          (iterate car (- k))))))
    (else
      (o (walk (car ks)) (apply walk (cdr ks))))))

(define (walk* . ks)
  (apply walk (reverse ks)))

) ; module tree-walkers

;(import tree-walkers simple-tests)
;(check ((flat '(0 1 2 3))
;        (deep '(0 (1 (2) 3) . 4)))
;  (iterate add1 5 1)
;  6
;  (iterate cdr 2 flat)
;  '(2 3)
;  ((walk -1 2) flat)
;  2
;  (caddr flat)
;  2
;  ((walk 2) flat)
;  '(2 3)
;  (cddr flat)
;  '(2 3)
;  ((walk* 1 -1 1 -2) deep)
;  2
;  ((walk 2) deep)
;  4
;  )

